// Copyright (C) 1998 Kazuhisa Iizuka

import java.io.*;
import java.util.*;

public class Data {
  static final int INTEGER = 0;
  static final int STRING  = 1;
  static final int ATOM    = 2;
  static final int FUNCTOR = 3;

  private int type;
  private int num;
  private String str;  // STRING, ATOM, ե󥯥̾
  private Data args[];

  /**
   * Ѥ륳󥹥ȥ饯.
   * ѤϤʤ
   * 󥹥󥹤ϡData.Integer(), Data.String(),
   *  Data.Atom(), Data.Functor() Ѥ롣
   */
  private Data(int type) {
    this.type = type;
  }

  protected Data(Data d) {
    this.type = d.type;
    
    switch (d.type) {
    case Data.INTEGER :
      this.num = d.num;
      break;
    case Data.STRING :
      this.str = d.str;
      break;
    case Data.ATOM :
      this.str = d.str;
      break;
    case Data.FUNCTOR :
      this.str = d.str;
      this.args = d.args;
      break;
    }
  }

  public static Data clone(Data d) {
    switch (d.type) {
    case Data.INTEGER :
      return Data.Integer(d.num);
    case Data.STRING :
      return Data.String(d.str);
    case Data.ATOM :
      return Data.Atom(d.str);
    case Data.FUNCTOR :
      Data args[] = new Data[d.args.length];
      for (int i = 0; i < d.args.length; i++) {
	args[i] = Data.clone(d.args[i]);
      }
      return Data.Functor(d.str, args);
    }
    
    return null;  // not reached
  }

  protected void replace(Data d) {
    this.type = d.type;
    this.num = d.num;
    this.str = d.str;
    this.args = d.args;
  }

  public int getDataType() {
    return type;
  }

  public static Data Integer(int num) {
    Data d = new Data(INTEGER);
    d.num = num;
    return d;
  }

  int getInteger() {
    return num;
  }

  public static Data String(String str) {
    Data d = new Data(STRING);
    d.str = new String(str);
    return d;
  }

  public String getString() {
    return new String(str);
  }

  public static Data Atom(String atom) {
    Data d = new Data(ATOM);
    d.str = new String(atom);
    return d;
  }

  public String getAtomName() {
    return new String(str);
  }

  public static Data Functor(String atom, Data args[]) {
    // ĹΥå
    if (args.length == 0) {
      return Atom(atom);
    }

    // Ȥ̥ݥ󥿤Ǥʤå
    for (int i = 0; i < args.length; i++) {
      if (args[i] == null)
	return Atom(atom);
    }

    Data d = new Data(FUNCTOR);
    d.str = new String(atom);

    // 򿷤Ȥ򥳥ԡ
    d.args = new Data[args.length];
    System.arraycopy(args, 0, d.args, 0, args.length);

    return d;
  }

  public String getFunctorName() {
    return new String(str);
  }

  public Data[] getFunctorArguments() {
    Data ret[] = new Data[args.length];
    System.arraycopy(args, 0, ret, 0, args.length);
    return ret;
  }

  static Data Nil() {
    return Data.Atom("[]");
  }

  public static Data Cons(Data car, Data cdr) {
    Data[] cons = new Data[] {car, cdr};
    return Data.Functor(".", cons);
  }

  public static Data[] ListToArray(Data list) {
    // ⤷[] ʤĹ֤
    if (Data.isNil(list))
      return new Data[0];

    if (Data.isCons(list)) {      // ꥹȤʤ ...
      Vector v = new Vector();
      Data d;

      // ꥹȤγǤ٥¸
      for(d = list; Data.isCons(d); d = d.args[1]) {
	v.addElement(d.args[0]);
      }

      // Ǹ夬 [] ǤʤСΥǡɲä
      if (!isNil(d))
	v.addElement(d);

      // ٥Ѵ
      Data ret[] = new Data[v.size()];
      v.copyInto(ret);

      return ret;
    }

    // ⤷ꥹȤǤʤСΥǡͣǤȤꥹȤ֤
    return new Data[] { Data.Cons(list, Nil()) };
  }

  public static Data List(Data list[]) {
    return makeList(list, 0);
  }

  private static Data makeList(Data list[], int p) {
    if (p == list.length)
      return Nil();
    else
      return Cons(list[p], makeList(list, p+1));
  }

  public static boolean isCons(Data d) {
    return    d.type == FUNCTOR
           && d.str.equals(".")
           && d.args.length == 2;
  }

  public static boolean isNil(Data d) {
    return d.type == ATOM && d.str.equals("[]");
  }


  /**
   * ʸ󤫤ǡ.
   */
  public static Data Parse(Reader in) throws IOException {
    return Data.ParseMain(in);
  }

  private static Data ParseMain(Reader in) throws IOException {
    char c = (char)in.read();

    switch (c) {
    case 'a' :
      for (int i = 1; i < "atom(".length(); i++)
	in.read();
      return parseAtom(in);
    case 'i' :
      for (int i = 1; i < "integer(".length(); i++)
	in.read();
      return parseInteger(in);
    case 'l' :
      for (int i = 1; i < "list(".length(); i++)
	in.read();
      return parseList(in);
    case 'f' :
      for (int i = 1; i < "functor(".length(); i++)
	in.read();
      return parseFunctor(in);
    case 's' :
      for (int i = 1; i < "string(".length(); i++)
	in.read();
      return parseString(in);
    }

    return null;
  }

  private static Data parseAtom(Reader in) throws IOException {
    // atom(...) | ex. atom([])
    char c;
    StringBuffer str = new StringBuffer();

    while ( (c = (char)in.read()) != ')' )
      str.append(c);

    return Data.Atom(str.toString());
  }

  private static Data parseInteger(Reader in) throws IOException {
    // integer(...) | ex. integer(-1) 

    char c;
    StringBuffer str = new StringBuffer();

    while ( (c = (char)in.read()) != ')' )
      str.append(c);

    int num =  Integer.valueOf(str.toString()).intValue();
    return Data.Integer(num);
  }

  private static Data parseString(Reader in) throws IOException {
    // string("...") | ex. string("abc")

    in.read();  // read '"'

    char c;
    StringBuffer str = new StringBuffer();

    while ( (c = (char)in.read()) != '\"')
      str.append(c);

    in.read();  // read ')'

    return Data.String(str.toString());
  }

  private static Data parseList(Reader in) throws IOException {
    // list([..|..]) | ex. list([atom(a)|atom([])])

    in.read();  // read '['

    Data car, cdr;

    car = Data.ParseMain(in);

    in.read();  // read '|'

    cdr = Data.ParseMain(in);

    in.read();  // read ']'
    in.read();  // read ')'

    return Data.Cons(car, cdr);
  }

  private static Data parseFunctor(Reader in) throws IOException {
    // functor(func(..)) | ex. functor(.(atom(one), atom([])))

    char c;
    StringBuffer str = new StringBuffer();

    while ( (c = (char)in.read()) != '(' )  // '(' ޤǤե󥯥̾
      str.append(c);

    Vector args = new Vector();

    do {
      args.addElement(Data.ParseMain(in));
    } while (in.read() != ')');  // read separator (',' or ')')

    in.read();  // read ')'

    Data[] args_array = new Data[args.size()];
    args.copyInto((Object[])args_array);

    return Data.Functor(str.toString(), args_array);
  }



  /**
   * ǡ뤿,åפ줿ʸѴ.
   */
  public static String getWTString(Data d) {
    return wrapStr(d);
  }

  /**
   * ǡåפ줿ʸѴ.
   */
  private static String wrapStr(Data d) {
     switch (d.type) {
     case Data.INTEGER :
       return convertInt(d);
     case Data.STRING :
       return convertStr(d);
     case Data.ATOM :
       return convertAtom(d);
     case Data.FUNCTOR :
       if (Data.isCons(d))
	 return convertList(d);
       else
	 return convertFunctor(d);
     }

     return "";
  }

  private static String convertInt(Data d) {
    // integer(...) | ex. integer(-1) 
    return "integer(" + Data.toStringInt(d) + ")";
  }

  private static String convertStr(Data d) {
    // string("...") | ex. string("abc")
    return "string(" + Data.toStringStr(d) + ")";
  }

  private static String convertAtom(Data d) {
    // atom(...) | ex. atom([])
    return "atom(" + Data.toStringAtom(d) + ")";
  }

  private static String convertList(Data d) {
    // list([..|..]) | ex. list([atom(a)|atom([])])
    return "list([" + Data.wrapStr(d.args[0]) + "|"
                    + wrapStr(d.args[1]) + "])";
  }

  private static String convertFunctor(Data d) {
    // functor(func(..)) | ex. functor(.(atom(one), atom([])))
    StringBuffer str = new StringBuffer();

    str.append("functor(");

    str.append(Data.toStringAtom(d)).append("(");
    str.append(wrapStr(d.args[0]));
    for (int i = 1; i < d.args.length; i++) {
      str.append(",");
      str.append(wrapStr(d.args[i]));
    }
    str.append(")");

    str.append(")");

    return str.toString();
  }



  /**
   * ǡʸѴ.
   */
  static String toStr(Data d) {
    if (d == null)
      return null;

    switch (d.type) {
     case Data.INTEGER :
       return toStringInt(d);
     case Data.STRING :
       return toStringStr(d);
     case Data.ATOM :
       return toStringAtom(d);
     case Data.FUNCTOR :
       if (Data.isCons(d))
	 return toStringList(d);
       else
	 return toStringFunctor(d);
     }

     return "";
  }

  private static String toStringInt(Data d) {
    return String.valueOf(d.num);
  }

  private static String toStringStr(Data d) {
    return "\"" + d.str + "\"";
  }

  // atom ¾ functor μե󥯥ȤƤƤФ
  private static String toStringAtom(Data d) {
    return d.str;
  }

  private static String toStringList(Data d) {
    StringBuffer str = new StringBuffer();

    str.append("[");

    for (;;) {
      str.append(Data.toStr(d.args[0]));

      if (!isCons(d.args[1]))
	break;

      str.append(",");
      d = d.args[1];
    }

    if (!isNil(d.args[1]))
      str.append("|").append(Data.toStr(d.args[1]));

    str.append("]");

    return str.toString();
  }

  private static String toStringFunctor(Data d) {
    StringBuffer str = new StringBuffer();

    str.append(Data.toStringAtom(d)).append("(");
    str.append(Data.toStr(d.args[0]));
    for (int i = 1; i < d.args.length; i++) {
      str.append(",");
      str.append(Data.toStr(d.args[i]));
    }
    str.append(")");

    return str.toString();
  }
}
