// Copyright (C) 1998 Kazuhisa Iizuka

import java.net.*;
import java.io.*;
import java.lang.reflect.*;

public class JK {
  Reader in = null;
  Writer out = null;
  String errorMessage = null;
  ReadThread rt = null;
  InputDataQueue queue;
  TagManager tag_m;
  ArgsManager arg_m;

  JK(InputStream in, OutputStream out) {
    this.in = new BufferedReader(new InputStreamReader(in));
    this.out = new OutputStreamWriter(out);
    queue = new InputDataQueue();
    rt = new ReadThread();
    rt.start();
    tag_m = new TagManager();
    arg_m = new ArgsManager(tag_m);
  }


  /**
   * ǡν񤭹.
   */
  void put(Data d) {
    put(0, d);
  }

  void put(int port, Data d) {
    if (out == null) {
      errorOccured("put : socket has already closed");
      return;
    }

    if (port < 0) {
      errorOccured("put : port is negative");
      return;
    }

    try {
      String send = new String();
      send += "data(";
      send += Data.getWTString(Data.Integer(port));
      send += ",";
      send += Data.getWTString(d);
      send += ")\n";             // \n : ȥåԥ󥰥

      out.write(send, 0, send.length());
      out.flush();
    } catch(Exception e) {
      errorOccured("put : Error at socket(" + e + ")");
      return;
    }
  }

  Data get() {
    return queue.remove();
  }

  Data call(String module_name, String goal_name, Data args[]) {
    Data args_list = Data.List(args);
    RetObject ret = arg_m.newObject();

    try {
      String send = new String();
      send += "call(";
      send += Data.getWTString(Data.Atom(module_name));
      send += ",";
      send += Data.getWTString(Data.Atom(goal_name));
      send += ",";
      send += Data.getWTString(args_list);
      send += ",";
      send += Data.getWTString(Data.Integer(ret.tag));
      send += ")\n";
      out.write(send, 0, send.length());
      out.flush();
    } catch(Exception e) {
      errorOccured("call : Error at socket(" + e + ")");
      return null;
    }

    Data return_value = ret.read();
    return return_value;
  }

  class ReadThread extends Thread {
    public void run() {
      try {
	for (;;) {
	  try {
	    int ic = in.read();
	    if (ic == -1) {
	      queue.finish();
	      close();
	      return;
	    }

	    char c = (char)ic;
	    if (c == 'r') {
	      // ret(Data,Tag)
	      in.read(); // e
	      in.read(); // t
	      in.read(); // (
	      Data data = Data.Parse(in);
	      in.read(); // ,
	      Data tag_d = Data.Parse(in);
	      in.read(); // )
	      in.read(); // \n
	      if (data == null || tag_d == null) {
		queue.finish();
		close();
		return;
	      }
	      int tag = tag_d.getInteger();
	      arg_m.setData(data, tag);  // ͤꤹ
	    } else if (c == 'c') {
	      // close or call
	      int nic = in.read();
	      if (nic == -1) {
		queue.finish();
		close();
		return;
	      }
	      char nc = (char)nic;
	      if (nc == 'l') {
		in.read(); // o
		in.read(); // s
		in.read(); // e
		in.read(); // \n
		queue.finish();
		close();
		return;
	      } else if (nc == 'a') {
		in.read(); // l
		in.read(); // l
		in.read(); // (
		Data cla = Data.Parse(in);
		in.read(); // ,
		Data method = Data.Parse(in);
		in.read(); // ,
		Data args = Data.Parse(in);
		in.read(); // ,
		Data tag = Data.Parse(in);
		in.read(); // )
		in.read(); // \n
		if (   cla == null || method == null
		     || args == null || tag == null) {
		  queue.finish();
		  close();
		  return;
		}
		procedureCall(cla, method, args, tag);
	      }
	    } else if (c == 'd') {
	      // data(Data)
	      in.read(); // a
	      in.read(); // t
	      in.read(); // a
	      in.read(); // (
	      Data d = Data.Parse(in);
	      in.read(); // )
	      in.read(); // \n

	      queue.add(d);
	    }
	  } catch(Exception e) {
	    errorOccured("ReadThread : Error at socket(" + e + ")");
	    return;
	  }
	}
      } catch (ThreadDeath d) {
	queue.finish();
	try {
	  if (in != null) {
	    in.close();
	    in = null;
	  }
	  if (out != null) {
	    out.close();
	    out = null;
	  }
	} catch(Exception e) {
	  // no error operation
	}
	throw d;
      }
    }
  }

  private void procedureCall(Data cl, Data me, Data ar, Data ta) {
    Class cla = null;
    Method method = null;
    Data args_d[] = Data.ListToArray(ar);
    Object args[] = new Object[args_d.length];
    Class args_class[] = new Class[args_d.length];
    Class data_class = ar.getClass();
    for (int i = 0; i < args.length; i++) {
      args[i] = args_d[i];
      args_class[i] = data_class;
    }
    int tag = ta.getInteger();
    String error_message = null;

    try {
      cla = Class.forName(cl.getString());
      method = cla.getMethod(me.getString(), args_class);
    } catch (ClassNotFoundException e) {
      error_message = e.toString();
    } catch (NoSuchMethodException e) {
      error_message = e.toString();
    } catch(SecurityException e) {
      error_message = e.toString();
    }
    if (   cla == null || method == null
	|| method.getReturnType() != data_class) {
      Data ret = Data.Functor("abnormal",
			      new Data[] {Data.String(error_message)});
      procedureCallReturn(ret, tag);
    } else {
      ProcedureCallThread th = new ProcedureCallThread(method, args, tag);
      th.start();
    }
  }  

  class ProcedureCallThread extends Thread {
    Method method;
    Object args[];
    int tag;

    ProcedureCallThread(Method method, Object args[], int tag) {
      this.method = method;
      this.args = args;
      this.tag = tag;
    }

    public void run() {
      Data re = null;
      String error_message = null;

      try {
	re = (Data) method.invoke(null, args);
      } catch(IllegalAccessException e) {
	error_message = e.toString();
      } catch(IllegalArgumentException e) {
	error_message = e.toString();
      } catch(InvocationTargetException e) {
	error_message = e.toString();
      } catch(NullPointerException e) {
	error_message = e.toString();
      }
      if (re == null) {
	Data ret = Data.Functor("abnormal",
				new Data[] {Data.String(error_message)});
	procedureCallReturn(ret, tag);
      } else {
	Data ret = Data.Functor("normal", new Data[] {re});
	procedureCallReturn(ret, tag);
      }
    }
  }

  private void procedureCallReturn(Data d, int tag) {
    try {
      String send = new String();
      send += "ret(";
      send += Data.getWTString(d);
      send += ",";
      send += Data.getWTString(Data.Integer(tag));
      send += ")\n";             // \n : ȥåԥ󥰥

      out.write(send, 0, send.length());
      out.flush();
    } catch(Exception e) {
      errorOccured("ret : Error at socket(" + e + ")");
    }
  }

  /**
   * 顼ȯνԤʤ.
   */
  private void errorOccured(String str) {
    // 顼å֤
    errorMessage = str;
    System.err.println(str);

    close();
  }

  public void close() {
    // ɤ߹ߥåɤߤ롣
    if (rt != null) {
      if (rt.isAlive())
	rt.stop();
      rt = null;
    }

    // ϥȥ꡼Ĥ
    try {
      if (in != null) {
	in.close();
	in = null;
      }
      if (out != null) {
	out.close();
	out = null;
      }
    } catch(Exception e) {
      // no error operation
    }
  }
}


class QueueElement {
  Data data;
  QueueElement next;

  QueueElement(Data d) { data = d; next = null; }
  void setNext(QueueElement n) { next = n; }
  QueueElement getNext() { return next; }
  Data getData() { return data; }
}


class InputDataQueue {
  /**
   * 塼򤢤魯ꥹȤκǽ.
   */
  private volatile QueueElement head;

  /**
   * 塼򤢤魯ꥹȤκǸ.
   */
  private QueueElement tail;

  /**
   * ǡϤλɤΥե饰.
   */
  private volatile boolean data_finish;

  InputDataQueue() {
    head = null;
    tail = null;
    data_finish = false;
  }

  /**
   * 塼ǡФ.
   * @return Фǡnull ξϡϤ̵򼨤
   */
  synchronized Data remove() {
    // Хåե˥ǡ̵ϡǡԤ
    if (head == null) {
      if (data_finish)
	return null;

      try {
	// add()  finish()  notify() Ԥ
	wait();
      } catch (InterruptedException ie) {
	// nothing to do
      }

      if (data_finish)
	return null;
    }

    QueueElement qe = head;

    head = qe.getNext();    // ꥹȤƬΰ֤򹹿

    return qe.getData();    // ꥹȤƬˤäǡ֤
  }

  synchronized void finish() {
    data_finish = true;
    notify();   // λ remove()  wait() 
  }

  synchronized void add(Data d) {
    if (d == null)
      return;

    // ǡꥹȤǤ
    QueueElement qe = new QueueElement(d);

    if (head == null)
      head = qe;         // ꥹȤƬ
    else
      tail.setNext(qe);  // ꥹȤ˺Ǹɲ

    tail = qe;  // ꥹȤκǸΰ֤ѹ

    notify();   // ǡɲä remove()  wait() 
  }
}
